/*************************************************************************
 *
 * DO NOT ALTER OR REMOVE COPYRIGHT NOTICES OR THIS FILE HEADER.
 * 
 * Copyright 2008 by Nakata, Maho
 *
 * MPACK - multiple precision arithmetic library
 *
 * This file is part of MPACK.
 *
 * MPACK is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License version 3
 * only, as published by the Free Software Foundation.
 *
 * MPACK is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Lesser General Public License version 3 for more details
 * (a copy is included in the LICENSE file that accompanied this code).
 *
 * You should have received a copy of the GNU Lesser General Public License
 * version 3 along with MPACK.  If not, see
 * <http://www.gnu.org/licenses/lgpl.html>
 * for a copy of the LGPLv3 License.
 *
 ************************************************************************/
/*
Copyright (c) 1992-2007 The University of Tennessee.  All rights reserved.

$COPYRIGHT$

Additional copyrights may follow

$HEADER$

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions are
met:

- Redistributions of source code must retain the above copyright
  notice, this list of conditions and the following disclaimer. 
  
- Redistributions in binary form must reproduce the above copyright
  notice, this list of conditions and the following disclaimer listed
  in this license in the documentation and/or other materials
  provided with the distribution.
  
- Neither the name of the copyright holders nor the names of its
  contributors may be used to endorse or promote products derived from
  this software without specific prior written permission.
  
THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
"AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT  
LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT 
OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT  
(INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE. 
*/

/*
Based on http://www.netlib.org/blas/cherk.f
Cherk performs one of the hermitian rank k operations
 C := alpha*A*conjg(A') + beta*C,
or
 C := alpha*conjg(A')*A + beta*C,
where  alpha and beta are real scalars, C is an n by n hermitian
matrix and A is an n by k matrix in the first case and a k by n
matrix in the second case.
*/

#include <mblas.h>

void
Cherk(const char *uplo, const char *trans, int n, int k, mpf_class alpha,
    mpc_class * A, int lda, mpf_class beta, mpc_class * C, int ldc)
{
    int nrowa, upper, info;
    mpc_class Zero;
    mpf_class One = 1.0;
    mpf_class rtemp;
    mpc_class temp;

    Zero.re = 0.0;
    Zero.im = 0.0;

    //Test the input parameters.
    if (Mlsame(trans, "N"))
	nrowa = n;
    else
	nrowa = k;
    upper = Mlsame(uplo, "U");

    info = 0;
    if ((!upper) && (!Mlsame(uplo, "L")))
	info = 1;
    else if ((!Mlsame(trans, "N")) && (!Mlsame(trans, "C")))
	info = 2;
    else if (n < 0)
	info = 3;
    else if (k < 0)
	info = 4;
    else if (lda < max(1, nrowa))
	info = 7;
    else if (ldc < max(1, n))
	info = 10;
    if (info != 0) {
	Mxerbla("Cherk ", info);
	return;
    }
    //quick return if possible.
    if ((n == 0) || (((alpha == real(Zero)) || (k == 0)) && (beta == One)))
	return;

    //and when alpha==Zero.
    if (alpha == real(Zero)) {
	if (upper) {
	    if (beta == real(Zero)) {
		for (int j = 0; j < n; j++) {
		    for (int i = 0; i <= j; i++) {
			C[i + j * ldc] = real(Zero);
		    }
		}
	    } else {
		for (int j = 0; j < n; j++) {
		    for (int i = 0; i < j; i++) {
			C[i + j * ldc] = beta * C[i + j * ldc];
		    }
		    C[j + j * ldc] = real(beta * C[j + j * ldc]);
		}
	    }
	} else {
	    if (beta == real(Zero)) {
		for (int j = 0; j < n; j++) {
		    for (int i = j; i < n; i++) {
			C[i + j * ldc] = real(Zero);
		    }
		}
	    } else {
		for (int j = 0; j < n; j++) {
		    C[j + j * ldc] = beta * real(C[j + j * ldc]);
		    for (int i = j + 1; i < n; i++) {
			C[i + j * ldc] = beta * C[i + j * ldc];
		    }
		}
	    }
	}
	return;
    }
    //start the operations.
    if (Mlsame(trans, "N")) {
	//Form C := alpha*A*conjg(A') + beta*C.
	if (upper) {
	    for (int j = 0; j < n; j++) {
		if (beta == real(Zero)) {
		    for (int i = 0; i <= j; i++) {
			C[i + j * ldc] = real(Zero);
		    }
		} else if (beta != One) {
		    for (int i = 0; i < j; i++) {
			C[i + j * ldc] = beta * C[i + j * ldc];
		    }
		    C[j + j * ldc] = beta * real(C[j + j * ldc]);
		} else {
		    C[j + j * ldc] = real(C[j + j * ldc]);
		}
		for (int l = 0; l < k; l++) {
		    if (A[j + l * lda] != Zero) {
			temp = alpha * conjg(A[j + l * lda]);
			for (int i = 0; i < j; i++) {
			    C[i + j * ldc] =
				C[i + j * ldc] + temp * A[i + l * lda];
			}
			C[j + j * ldc] =
			    real(C[j + j * ldc]) + real(temp * A[j + l * lda]);
		    }
		}
	    }
	} else {
	    for (int j = 0; j < n; j++) {
		if (beta == real(Zero)) {
		    for (int i = j; i < n; i++) {
			C[i + j * ldc] = Zero;
		    }
		} else if (beta != One) {
		    C[j + j * ldc] = beta * real(C[j + j * ldc]);
		    for (int i = j + 1; i < n; i++) {
			C[i + j * ldc] = beta * C[i + j * ldc];
		    }
		} else {
		    C[j + j * ldc] = real(C[j + j * ldc]);
		}
		for (int l = 0; l < k; l++) {
		    if (A[j + l * lda] != Zero) {
			temp = alpha * conjg(A[j + l * lda]);
			C[j + j * ldc] =
			    real(C[j + j * ldc]) + real(temp * A[j + l * lda]);
			for (int i = j + 1; i < n; i++) {
			    C[i + j * ldc] =
				C[i + j * ldc] + temp * A[i + l * lda];
			}
		    }
		}
	    }
	}
    } else {
	//form  C := alpha*conjg(A')*A + beta*C.
	if (upper) {
	    for (int j = 0; j < n; j++) {
		for (int i = 0; i < j; i++) {
		    temp = Zero;
		    for (int l = 0; l < k; l++) {
			temp = temp + conjg(A[l + i * lda]) * A[l + j * lda];
		    }
		    if (beta == real(Zero)) {
			C[i + j * ldc] = alpha * temp;
		    } else {
			C[i + j * ldc] = alpha * temp + beta * C[i + j * ldc];
		    }
		}
		rtemp = real(Zero);
		for (int l = 0; l < k; l++) {
		    rtemp =
			rtemp + real(conjg(A[l + j * lda]) * A[l + j * lda]);
		}
		if (beta == real(Zero)) {
		    C[j + j * ldc] = alpha * rtemp;
		} else {
		    C[j + j * ldc] =
			alpha * rtemp + beta * real(C[j + j * ldc]);
		}
	    }
	} else {
	    for (int j = 0; j < n; j++) {
		rtemp = real(Zero);
		for (int l = 0; l < k; l++) {
		    rtemp =
			rtemp + real(conjg(A[l + j * lda]) * A[l + j * lda]);
		}
		if (beta == real(Zero)) {
		    C[j + j * ldc] = alpha * rtemp;
		} else {
		    C[j + j * ldc] =
			alpha * rtemp + beta * real(C[j + j * ldc]);
		}
		for (int i = j + 1; i < n; i++) {
		    temp = Zero;
		    for (int l = 0; l < k; l++)
			temp = temp + conjg(A[l + i * lda]) * A[l + j * lda];
		    if (beta == real(Zero)) {
			C[i + j * ldc] = alpha * temp;
		    } else {
			C[i + j * ldc] = alpha * temp + beta * C[i + j * ldc];
		    }
		}
	    }
	}
    }
    return;
}
