/*************************************************************************
 *
 * DO NOT ALTER OR REMOVE COPYRIGHT NOTICES OR THIS FILE HEADER.
 * 
 * Copyright 2008 by Nakata, Maho
 *
 * MPACK - multiple precision arithmetic library
 *
 * This file is part of MPACK.
 *
 * MPACK is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License version 3
 * only, as published by the Free Software Foundation.
 *
 * MPACK is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Lesser General Public License version 3 for more details
 * (a copy is included in the LICENSE file that accompanied this code).
 *
 * You should have received a copy of the GNU Lesser General Public License
 * version 3 along with MPACK.  If not, see
 * <http://www.gnu.org/licenses/lgpl.html>
 * for a copy of the LGPLv3 License.
 *
 ************************************************************************/
#include <stdio.h>
#include <iostream>
#include <fstream>
#include <cstdio>
#include <cstdlib>
#include <cmath>
#include <cstring>
#include <complex>

#include <mblas.h>
#include "blas.h"
#include "mblas.debug.h"

#define MIN_N 1
#define MAX_N 10
#define MAX_K 10
#define MIN_LDA 1
#define MAX_LDA 10
#define MIN_INCX -2
#define MAX_INCX 3
#define MIN_INCY -2
#define MAX_INCY 2
#define MAX_ITER 3

void
Chbmv_test3(const char *uplo, mpc_class alpha, mpc_class beta)
{
    int errorflag = FALSE;

    for (int incx = MIN_INCX; incx <= MAX_INCX; incx++) {
	for (int incy = MIN_INCY; incy < MAX_INCY; incy++) {
	    for (int n = MIN_N; n < MAX_N; n++) {
		for (int k = 0; k < MAX_K; k++) {
		    for (int lda = k + 1; lda < MAX_LDA; lda++) {

			mpc_class *A = new mpc_class[MAT_A(lda, n)];
			mpc_class *x = new mpc_class[VEC_A(n, incx)];
			mpc_class *y = new mpc_class[VEC_A(n, incy)];

			complex < double >*Ad =
			    new complex < double >[MAT_A(lda, n)];
			complex < double >*xd =
			    new complex < double >[VEC_A(n, incx)];
			complex < double >*yd =
			    new complex < double >[VEC_A(n, incy)];

			complex < double >alphad, betad;

			alphad.real() = alpha.re.get_d();
			alphad.imag() = alpha.im.get_d();

			betad.real() = beta.re.get_d();
			betad.imag() = beta.im.get_d();

			for (int iter = 0; iter < MAX_ITER; iter++) {
			    set_random_vector(A, Ad, MAT_A(lda, n));
			    set_random_vector(x, xd, VEC_A(n, incx));
			    set_random_vector(y, yd, VEC_A(n, incy));

			    mpack_errno = 0;
			    Chbmv(uplo, n, k, alpha, A, lda, x, incx, beta, y,
				incy);
			    zhbmv_f77(uplo, &n, &k, &alphad, Ad, &lda, xd,
				&incx, &betad, yd, &incy);
			    if (mpack_errno != 0) {
				break;
			    }

			    double diff;

			    diff = vec_diff(y, yd, VEC_A(n, incy), 1);
			    if (fabs(diff) > EPSILON) {
				printf("#error %lf!!\n", diff);
				errorflag = TRUE;
			    } else {
//                              printf("#okay\n");
			    }
			}
			delete[]Ad;
			delete[]xd;
			delete[]yd;
			delete[]x;
			delete[]y;
			delete[]A;
		    }
		}
	    }
	}
    }
    if (errorflag == TRUE) {
	printf("Chbmv test failed...\n");
	exit(-1);
    }
}

void
Chbmv_test2(const char *uplo)
{
    mpc_class alpha, beta;
    mpc_class Czero, Cone;

    Czero.re = 0.0;
    Czero.im = 0.0;
    Cone.re = 1.0;
    Cone.im = 0.0;

//alpha=*, beta=*
    alpha = mpc_randomnumber();
    beta = mpc_randomnumber();
    Chbmv_test3(uplo, alpha, beta);

//a=0, b=0;
    alpha = Czero;
    beta = Czero;
    Chbmv_test3(uplo, alpha, beta);

//a=1, b=0;
    alpha = Cone;
    beta = Czero;
    Chbmv_test3(uplo, alpha, beta);

//a=0, b=1;
    alpha = Czero;
    beta = Cone;
    Chbmv_test3(uplo, alpha, beta);

//a=1, b=1;
    alpha = Cone;
    beta = Cone;
    Chbmv_test3(uplo, alpha, beta);

//a=*, b=0;
    alpha = mpc_randomnumber();
    beta = Czero;
    Chbmv_test3(uplo, alpha, beta);

//a=*, b=1;
    alpha = mpc_randomnumber();
    beta = Cone;
    Chbmv_test3(uplo, alpha, beta);

//a=0, b=*;
    alpha = Czero;
    beta = mpc_randomnumber();
    Chbmv_test3(uplo, alpha, beta);

//a=1, b=*;
    alpha = Cone;
    beta = mpc_randomnumber();
    Chbmv_test3(uplo, alpha, beta);
}

void
Chbmv_test()
{
    Chbmv_test2("U");
    Chbmv_test2("L");
}


int
main(int argc, char *argv[])
{
//initialization
    int default_prec = 200;
    gmp_randstate_t state;

    mpf_set_default_prec(default_prec);
    uniformrandom = new gmp_randclass(gmp_randinit_default);
    Chbmv_test();
    printf("Chbmv test passed...\n");

}
