/*************************************************************************
 *
 * DO NOT ALTER OR REMOVE COPYRIGHT NOTICES OR THIS FILE HEADER.
 * 
 * Copyright 2008 by Nakata, Maho
 *
 * MPACK - multiple precision arithmetic library
 *
 * This file is part of MPACK.
 *
 * MPACK is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License version 3
 * only, as published by the Free Software Foundation.
 *
 * MPACK is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Lesser General Public License version 3 for more details
 * (a copy is included in the LICENSE file that accompanied this code).
 *
 * You should have received a copy of the GNU Lesser General Public License
 * version 3 along with MPACK.  If not, see
 * <http://www.gnu.org/licenses/lgpl.html>
 * for a copy of the LGPLv3 License.
 *
 ************************************************************************/
#include <stdio.h>
#include <iostream>
#include <fstream>
#include <cstdio>
#include <cstdlib>
#include <cmath>
#include <cstring>
#include <complex>

#include <mblas.h>
#include "blas.h"
#include "mblas.debug.h"

#define MIN_M 2
#define MAX_M 10
#define MIN_N 2
#define MAX_N 10
#define MIN_LDA 2
#define MAX_LDA 10
#define MIN_LDB 2
#define MAX_LDB 10
#define MIN_LDC 2
#define MAX_LDC 10
#define MAX_ITER 1

void
Chemm_test3(const char *side, const char *uplo, mpc_class alpha,
    mpc_class beta)
{
    int errorflag = FALSE;

    for (int n = MIN_N; n < MAX_N; n++) {
	for (int m = MIN_M; m < MAX_M; m++) {

	    int minlda;

	    if (Mlsame(side, "L"))
		minlda = max(1, m);
	    else
		minlda = max(1, n);

	    for (int lda = minlda; lda < MAX_LDA; lda++) {
		for (int ldb = max(1, m); ldb < MAX_LDB; ldb++) {
		    for (int ldc = max(1, m); ldc < MAX_LDC; ldc++) {
/*
			printf
			    ("#n is %d, m is %d, lda is %d, ldb is %d, ldc is %d ",
			    n, m, lda, ldb, ldc);
			printf("side is %s, uplo is %s \n", side, uplo);
*/
			mpc_class *A;
			mpc_class *B;
			mpc_class *C;
			complex < double >*Ad;
			complex < double >*Bd;
			complex < double >*Cd;

			int ka;

			if (Mlsame(side, "L"))
			    ka = m;
			else
			    ka = n;

			A = new mpc_class[MAT_A(lda, ka)];
			B = new mpc_class[MAT_A(ldb, n)];
			C = new mpc_class[MAT_A(ldc, n)];
			Ad = new complex < double >[MAT_A(lda, ka)];
			Bd = new complex < double >[MAT_A(ldb, n)];
			Cd = new complex < double >[MAT_A(ldc, n)];

			complex < double >alphad, betad;

			alphad.real() = alpha.re.get_d();
			alphad.imag() = alpha.im.get_d();

			betad.real() = beta.re.get_d();
			betad.imag() = beta.im.get_d();

			for (int iter = 0; iter < MAX_ITER; iter++) {
			    set_random_vector(A, Ad, MAT_A(lda, ka));
			    set_random_vector(B, Bd, MAT_A(ldb, n));
			    set_random_vector(C, Cd, MAT_A(ldc, n));

			    mpack_errno = 0;
			    Chemm(side, uplo, m, n, alpha, A, lda, B, ldb,
				beta, C, ldc);
			    zhemm_f77(side, uplo, &m, &n, &alphad, Ad, &lda,
				Bd, &ldb, &betad, Cd, &ldc);
			    if (mpack_errno != 0) {
				break;
			    }

			    double diff;

			    diff = vec_diff(C, Cd, MAT_A(ldc, n), 1);
			    if (fabs(diff) > EPSILON) {
				printf("#error %lf!!\n", diff);
				errorflag = TRUE;
			    } else {
//                              printf("#okay\n");
			    }
			}
			delete[]Cd;
			delete[]Bd;
			delete[]Ad;

			delete[]C;
			delete[]B;
			delete[]A;
		    }
		}
	    }
	}
    }
    if (errorflag == TRUE) {
	printf("Chemm test failed...\n");
	exit(-1);
    }

}

void
Chemm_test2(const char *side, const char *uplo)
{
    mpc_class alpha, beta;
    mpc_class Czero, Cone;

    Czero.re = 0.0;
    Czero.im = 0.0;
    Cone.re = 1.0;
    Cone.im = 0.0;

//alpha=*, beta=*
    alpha = mpc_randomnumber();
    beta = mpc_randomnumber();
    Chemm_test3(side, uplo, alpha, beta);

//a=0, b=0;
    alpha = Czero;
    beta = Czero;
    Chemm_test3(side, uplo, alpha, beta);

//a=1, b=0;
    alpha = Cone;
    beta = Czero;
    Chemm_test3(side, uplo, alpha, beta);

//a=0, b=1;
    alpha = Czero;
    beta = Cone;
    Chemm_test3(side, uplo, alpha, beta);

//a=1, b=1;
    alpha = Cone;
    beta = Cone;
    Chemm_test3(side, uplo, alpha, beta);

//a=*, b=0;
    alpha = mpc_randomnumber();
    beta = Czero;
    Chemm_test3(side, uplo, alpha, beta);

//a=*, b=1;
    alpha = mpc_randomnumber();
    beta = Cone;
    Chemm_test3(side, uplo, alpha, beta);

//a=0, b=*;
    alpha = Czero;
    beta = mpc_randomnumber();
    Chemm_test3(side, uplo, alpha, beta);

//a=1, b=*;
    alpha = Cone;
    beta = mpc_randomnumber();
    Chemm_test3(side, uplo, alpha, beta);
}

void
Chemm_test()
{

    Chemm_test2("L", "U");
    Chemm_test2("L", "L");
    Chemm_test2("R", "U");
    Chemm_test2("R", "L");

}


int
main(int argc, char *argv[])
{
//initialization
    int default_prec = 200;
    gmp_randstate_t state;

    mpf_set_default_prec(default_prec);
    uniformrandom = new gmp_randclass(gmp_randinit_default);
    VERBOSE = 1;

    Chemm_test();
    printf("Chemm test passed...\n");
}
