/*************************************************************************
 *
 * DO NOT ALTER OR REMOVE COPYRIGHT NOTICES OR THIS FILE HEADER.
 * 
 * Copyright 2008 by Nakata, Maho
 *
 * MPACK - multiple precision arithmetic library
 *
 * This file is part of MPACK.
 *
 * MPACK is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License version 3
 * only, as published by the Free Software Foundation.
 *
 * MPACK is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Lesser General Public License version 3 for more details
 * (a copy is included in the LICENSE file that accompanied this code).
 *
 * You should have received a copy of the GNU Lesser General Public License
 * version 3 along with MPACK.  If not, see
 * <http://www.gnu.org/licenses/lgpl.html>
 * for a copy of the LGPLv3 License.
 *
 ************************************************************************/
#include <stdio.h>
#include <iostream>
#include <fstream>
#include <cstdio>
#include <cstdlib>
#include <cmath>
#include <cstring>
#include <complex>

#include <mblas.h>
#include "blas.h"
#include "mblas.debug.h"

#define MIN_K 1
#define MAX_K 12
#define MIN_N 1
#define MAX_N 12
#define MIN_LDA 2
#define MAX_LDA 12
#define MIN_LDC 2
#define MAX_LDC 12
#define MAX_ITER 1

void
Csyrk_test3(const char *uplo, const char *trans, mpc_class alpha,
    mpc_class beta)
{
    int errorflag = FALSE;

    for (int n = MIN_N; n < MAX_N; n++) {
	for (int k = MIN_K; k < MAX_K; k++) {

	    int minlda;

	    if (Mlsame(trans, "N"))
		minlda = max(1, n);
	    else
		minlda = max(1, k);

	    for (int lda = minlda; lda < MAX_LDA; lda++) {
		for (int ldc = max(1, n); ldc < MAX_LDC; ldc++) {

/*		    printf("#n is %d, k is %d, lda is %d, ldc is %d\n", n, k,
			lda, ldc);
		    printf("#uplo is %s, trans is %s\n", uplo, trans); */

		    mpc_class *A;
		    mpc_class *C;
		    complex < double >*Ad;
		    complex < double >*Cd;

		    int ka;

		    if (Mlsame(trans, "N"))
			ka = k;
		    else
			ka = n;

		    A = new mpc_class[MAT_A(lda, ka)];
		    C = new mpc_class[MAT_A(ldc, n)];
		    Ad = new complex < double >[MAT_A(lda, ka)];
		    Cd = new complex < double >[MAT_A(ldc, n)];

		    complex < double >alphad, betad;

		    alphad.real() = alpha.re.get_d();
		    alphad.imag() = alpha.im.get_d();
		    betad.real() = beta.re.get_d();
		    betad.imag() = beta.im.get_d();

		    for (int iter = 0; iter < MAX_ITER; iter++) {
			set_random_vector(A, Ad, MAT_A(lda, ka));
			set_random_vector(C, Cd, MAT_A(ldc, n));

			Csyrk(uplo, trans, n, k, alpha, A, lda, beta, C, ldc);
			zsyrk_f77(uplo, trans, &n, &k, &alphad, Ad, &lda,
			    &betad, Cd, &ldc);
			if (mpack_errno != 0) {
			    break;
			}
			double diff;

			diff = vec_diff(C, Cd, MAT_A(ldc, n), 1);
			if (fabs(diff) > EPSILON) {
			    printf("#error %lf!!\n", diff);
			    errorflag = TRUE;
			} else {
//                          printf("#okay\n");
			}
		    }
		    delete[]Cd;
		    delete[]Ad;

		    delete[]C;
		    delete[]A;
		}
	    }
	}
    }

    if (errorflag == TRUE) {
	printf("Csyrk test failed...\n");
	exit(-1);
    }
}

void
Csyrk_test2(const char *side, const char *uplo)
{
    mpc_class alpha, beta;
    mpc_class Czero, Cone;

    Czero.re = 0.0;
    Czero.im = 0.0;
    Cone.re = 1.0;
    Cone.im = 0.0;

//a=*, b=*;
    alpha = mpc_randomnumber();
    beta = mpc_randomnumber();
    Csyrk_test3(side, uplo, alpha, beta);

//a=*, b=1;
    alpha = mpc_randomnumber();
    beta = Cone;
    Csyrk_test3(side, uplo, alpha, beta);

//a=1, b=*;
    alpha = Cone;
    beta = mpc_randomnumber();
    Csyrk_test3(side, uplo, alpha, beta);

//a=*, b=0;
    alpha = mpc_randomnumber();
    beta = Czero;
    Csyrk_test3(side, uplo, alpha, beta);

//a=1, b=0;
    alpha = Cone;
    beta = Czero;
    Csyrk_test3(side, uplo, alpha, beta);

//a=1, b=1;
    alpha = Cone;
    beta = Cone;
    Csyrk_test3(side, uplo, alpha, beta);

//a=0, b=*;
    alpha = Czero;
    beta = mpc_randomnumber();
    Csyrk_test3(side, uplo, alpha, beta);

//a=0, b=0;
    alpha = Czero;
    beta = Czero;
    Csyrk_test3(side, uplo, alpha, beta);

//a=0, b=1;
    alpha = Czero;
    beta = Cone;
    Csyrk_test3(side, uplo, alpha, beta);

}

void
Csyrk_test()
{
    Csyrk_test2("U", "N");
    Csyrk_test2("L", "N");
    Csyrk_test2("U", "C");
    Csyrk_test2("L", "C");
}


int
main(int argc, char *argv[])
{
//initialization
    int default_prec = 200;
    gmp_randstate_t state;

    mpf_set_default_prec(default_prec);
    uniformrandom = new gmp_randclass(gmp_randinit_default);
    VERBOSE = 1;

    Csyrk_test();
    printf("Csyrk test passed...\n");
}
