/*************************************************************************
 *
 * DO NOT ALTER OR REMOVE COPYRIGHT NOTICES OR THIS FILE HEADER.
 * 
 * Copyright 2008 by Nakata, Maho
 *
 * MPACK - multiple precision arithmetic library
 *
 * This file is part of MPACK.
 *
 * MPACK is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License version 3
 * only, as published by the Free Software Foundation.
 *
 * MPACK is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Lesser General Public License version 3 for more details
 * (a copy is included in the LICENSE file that accompanied this code).
 *
 * You should have received a copy of the GNU Lesser General Public License
 * version 3 along with MPACK.  If not, see
 * <http://www.gnu.org/licenses/lgpl.html>
 * for a copy of the LGPLv3 License.
 *
 ************************************************************************/
#include <stdio.h>
#include <iostream>
#include <fstream>
#include <cstdio>
#include <cstdlib>
#include <cmath>
#include <cstring>
#include <complex>

#include <mblas.h>
#include "blas.h"
#include "mblas.debug.h"

#define MIN_M 2
#define MAX_M 10
#define MIN_N 2
#define MAX_N 10
#define MIN_K 2
#define MAX_K 10
#define MIN_LDA 2
#define MAX_LDA 10
#define MIN_LDB 2
#define MAX_LDB 10
#define MIN_LDC 2
#define MAX_LDC 10
#define MAX_ITER 1

void
Rgemm_test3(const char *transa, const char *transb, mpf_class alpha,
    mpf_class beta)
{
    int errorflag = FALSE;

    for (int k = MIN_K; k < MAX_K; k++) {
	for (int n = MIN_N; n < MAX_N; n++) {
	    for (int m = MIN_M; m < MAX_M; m++) {

		int minlda, minldb;

		if (Mlsame(transa, "N"))
		    minlda = max(1, m);
		else
		    minlda = max(1, k);

		if (Mlsame(transb, "N"))
		    minldb = max(1, k);
		else
		    minldb = max(1, n);

		for (int lda = minlda; lda < MAX_LDA; lda++) {
		    for (int ldb = minldb; ldb < MAX_LDB; ldb++) {
			for (int ldc = max(1, m); ldc < MAX_LDC; ldc++) {
/*
			    printf
				("#k is %d, n is %d, m is %d, lda is %d, ldb is %d, ldc is %d\n",
				k, n, m, lda, ldb, ldc);
			    printf("#transa is %s, transb is %s\n", transa,
				   transb);
*/

			    mpf_class *A;
			    mpf_class *B;
			    mpf_class *C;
			    double *Ad;
			    double *Bd;
			    double *Cd;

			    int ka, kb, kc;

			    if (Mlsame(transa, "N"))
				ka = k;
			    else
				ka = m;

			    if (Mlsame(transb, "N"))
				kb = n;
			    else
				kb = k;

			    A = new mpf_class[MAT_A(lda, ka)];
			    B = new mpf_class[MAT_A(ldb, kb)];
			    C = new mpf_class[MAT_A(ldc, n)];
			    Ad = new double[MAT_A(lda, ka)];
			    Bd = new double[MAT_A(ldb, kb)];
			    Cd = new double[MAT_A(ldc, n)];

			    double alphad, betad;

			    alphad = alpha.get_d();
			    betad = beta.get_d();

			    for (int iter = 0; iter < MAX_ITER; iter++) {
				set_random_vector(A, Ad, MAT_A(lda, ka));
				set_random_vector(B, Bd, MAT_A(ldb, kb));
				set_random_vector(C, Cd, MAT_A(ldc, n));

				mpack_errno = 0;
				Rgemm(transa, transb, m, n, k, alpha, A, lda,
				    B, ldb, beta, C, ldc);
				dgemm_f77(transa, transb, &m, &n, &k, &alphad,
				    Ad, &lda, Bd, &ldb, &betad, Cd, &ldc);
				if (mpack_errno != 0) {
				    break;
				}

				double diff;

				diff = vec_diff(C, Cd, MAT_A(ldc, n), 1);
				if (fabs(diff) > EPSILON) {
				    printf("#error %lf!!\n", diff);
				    errorflag = TRUE;
				} else {
//                                  printf("#okay\n");
				}
			    }
			    delete[]Cd;
			    delete[]Bd;
			    delete[]Ad;

			    delete[]C;
			    delete[]B;
			    delete[]A;
			}
		    }
		}
	    }
	}
    }

    if (errorflag == TRUE) {
	printf("Rgemm test failed...\n");
	exit(-1);
    }
}

void
Rgemm_test2(const char *uplo, const char *trans)
{
    mpf_class alpha;
    mpf_class beta;
    mpf_class Rzero, Rone;

    Rzero = 0.0;
    Rone = 1.0;

//a=0, b=*;
    alpha = Rzero;
    beta = mpf_randomnumber();
    Rgemm_test3(uplo, trans, alpha, beta);

//a=*, b=0;
    alpha = mpf_randomnumber();
    beta = Rzero;
    Rgemm_test3(uplo, trans, alpha, beta);

//a=*, b=1;
    alpha = mpf_randomnumber();
    beta = Rone;
    Rgemm_test3(uplo, trans, alpha, beta);

//alpha=*, beta=*
    alpha = mpf_randomnumber();
    beta = mpf_randomnumber();
    Rgemm_test3(uplo, trans, alpha, beta);

//a=0, b=0;
    alpha = Rzero;
    beta = Rzero;
    Rgemm_test3(uplo, trans, alpha, beta);

//a=0, b=1;
    alpha = Rzero;
    beta = Rone;
    Rgemm_test3(uplo, trans, alpha, beta);

//a=1, b=0;
    alpha = Rone;
    beta = Rzero;
    Rgemm_test3(uplo, trans, alpha, beta);

//a=1, b=1;
    alpha = Rone;
    beta = Rone;
    Rgemm_test3(uplo, trans, alpha, beta);

//a=1, b=*;
    alpha = Rone;
    beta = mpf_randomnumber();
    Rgemm_test3(uplo, trans, alpha, beta);
}

int
main(int argc, char *argv[])
{
//initialization
    int default_prec = 100;
    gmp_randstate_t state;

    mpf_set_default_prec(default_prec);
    uniformrandom = new gmp_randclass(gmp_randinit_default);
    VERBOSE = 0;

    Rgemm_test2("N", "N");
    Rgemm_test2("N", "T");
    Rgemm_test2("N", "C");

    Rgemm_test2("T", "N");
    Rgemm_test2("T", "C");
    Rgemm_test2("T", "T");

    Rgemm_test2("C", "N");
    Rgemm_test2("C", "T");
    Rgemm_test2("C", "C");
    printf("Rgemm test passed...\n");

}
