/*************************************************************************
 *
 * DO NOT ALTER OR REMOVE COPYRIGHT NOTICES OR THIS FILE HEADER.
 * 
 * Copyright 2008 by Nakata, Maho
 *
 * MPACK - multiple precision arithmetic library
 *
 * This file is part of MPACK.
 *
 * MPACK is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License version 3
 * only, as published by the Free Software Foundation.
 *
 * MPACK is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Lesser General Public License version 3 for more details
 * (a copy is included in the LICENSE file that accompanied this code).
 *
 * You should have received a copy of the GNU Lesser General Public License
 * version 3 along with MPACK.  If not, see
 * <http://www.gnu.org/licenses/lgpl.html>
 * for a copy of the LGPLv3 License.
 *
 ************************************************************************/
#include <stdio.h>
#include <iostream>
#include <fstream>
#include <cstdio>
#include <cstdlib>
#include <cmath>
#include <cstring>
#include <complex>

#include <mblas.h>
#include "blas.h"
#include "mblas.debug.h"

#define MIN_N 1
#define MAX_N 20
#define MIN_M 1
#define MAX_M 20
#define MIN_LDA 2
#define MAX_LDA 20
#define MIN_LDB 2
#define MAX_LDB 20
#define MAX_ITER 1

void
Rtrsm_test3(const char *side, const char *uplo, const char *transa,
    const char *diag, mpf_class alpha)
{
    int errorflag = FALSE;

    for (int m = MIN_M; m < MAX_M; m++) {
	for (int n = MIN_N; n < MAX_N; n++) {
	    int minlda;

	    if (Mlsame(side, "L"))
		minlda = max(1, m);
	    else
		minlda = max(1, n);

	    for (int lda = minlda; lda < MAX_LDA; lda++) {
		for (int ldb = max(1, m); ldb < MAX_LDB; ldb++) {

		    int k;

		    if (Mlsame(side, "L"))
			k = m;
		    if (Mlsame(side, "R"))
			k = n;

		    mpf_class *A;
		    mpf_class *B;
		    double *Ad;
		    double *Bd;

		    A = new mpf_class[MAT_A(lda, k)];
		    B = new mpf_class[MAT_A(ldb, n)];
		    Ad = new double[MAT_A(lda, k)];
		    Bd = new double[MAT_A(ldb, n)];

		    double alphad;

		    alphad = alpha.get_d();

		    for (int iter = 0; iter < MAX_ITER; iter++) {
			set_random_vector(A, Ad, MAT_A(lda, k));
			set_random_vector(B, Bd, MAT_A(ldb, n));

			Rtrsm(side, uplo, transa, diag, m, n, alpha, A, lda, B,
			    ldb);
			dtrsm_f77(side, uplo, transa, diag, &m, &n, &alphad,
			    Ad, &lda, Bd, &ldb);
			if (mpack_errno != 0) {
			    break;
			}
			double diff;

			diff = vec_diff(B, Bd, MAT_A(ldb, n), 1);
			if (fabs(diff) > EPSILON3) {
			    printf("#error %.16lf!!\n", diff);
			    errorflag = TRUE;
			    printf("#n is %d, m is %d, lda is %d, ldb is %d\n",
				n, m, lda, ldb);
			    printf("#side is %s, uplo is %s transa is %s\n",
				side, uplo, transa);

			} else {
//                          printf("#okay\n");
			}
		    }
		    delete[]Bd;
		    delete[]Ad;

		    delete[]B;
		    delete[]A;
		}
	    }
	}
    }

    if (errorflag == TRUE) {
	printf("Rtrsm test failed...\n");
	exit(-1);
    }
}

void
Rtrsm_test2(const char *side, const char *uplo, const char *transa,
    const char *diag)
{
    mpf_class alpha;
    mpf_class Rone, Rzero;

    Rzero = 0.0;
    Rone = 1.0;

//a=*
    alpha = mpf_randomnumber();
    Rtrsm_test3(side, uplo, transa, diag, alpha);

//a=1
    alpha = Rone;
    Rtrsm_test3(side, uplo, transa, diag, alpha);

//a=0
    alpha = Rzero;
    Rtrsm_test3(side, uplo, transa, diag, alpha);
}

void
Rtrsm_test()
{
    Rtrsm_test2("L", "U", "N", "U");
    Rtrsm_test2("L", "U", "N", "N");
    Rtrsm_test2("L", "U", "T", "U");
    Rtrsm_test2("L", "U", "T", "N");
    Rtrsm_test2("L", "U", "C", "U");
    Rtrsm_test2("L", "U", "C", "N");
    Rtrsm_test2("L", "L", "N", "U");
    Rtrsm_test2("L", "L", "N", "N");
    Rtrsm_test2("L", "L", "T", "U");
    Rtrsm_test2("L", "L", "T", "N");
    Rtrsm_test2("L", "L", "C", "U");
    Rtrsm_test2("L", "L", "C", "N");

    Rtrsm_test2("R", "U", "N", "U");
    Rtrsm_test2("R", "U", "N", "N");
    Rtrsm_test2("R", "U", "T", "U");
    Rtrsm_test2("R", "U", "T", "N");
    Rtrsm_test2("R", "U", "C", "U");
    Rtrsm_test2("R", "U", "C", "N");
    Rtrsm_test2("R", "L", "N", "U");
    Rtrsm_test2("R", "L", "N", "N");
    Rtrsm_test2("R", "L", "T", "U");
    Rtrsm_test2("R", "L", "T", "N");
    Rtrsm_test2("R", "L", "C", "U");
    Rtrsm_test2("R", "L", "C", "N");

}


int
main(int argc, char *argv[])
{
//initialization
    int default_prec = 200;
    gmp_randstate_t state;

    mpf_set_default_prec(default_prec);
    uniformrandom = new gmp_randclass(gmp_randinit_default);
    VERBOSE = 1;

    Rtrsm_test();
    printf("Rtrsm test passed...\n");
}
