/**********************************************/
/* efunc.c: Elemantary functions              */
/* Copyright (c) 2000- Tomonori Kouya         */
/*                                            */
/* This library is free software; you can re- */
/* distribute it and/or modify it under the   */
/* terms of the GNU Lesser General Public     */
/* License as published by the Free Software  */
/* Foundation; either version 2.1 of the      */
/* License, or (at your option) any later     */
/* version.                                   */
/*                                            */
/* This library is distributed in the hope    */
/* that it will be useful, but WITHOUT ANY    */
/* WARRANTY; without even the implied         */
/* warranty of MERCHANTABILITY or FITNESS FOR */
/* A PARTICULAR PURPOSE.  See the GNU Lesser  */
/* General Public License for more details.   */
/**********************************************/

/*************************************************************************
 *
 * DO NOT ALTER OR REMOVE COPYRIGHT NOTICES OR THIS FILE HEADER.
 * 
 * Copyright 2008 by Nakata, Maho
 *
 * MPACK - multiple precision arithmetic library
 *
 * This file is part of MPACK.
 *
 * MPACK is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License version 3
 * only, as published by the Free Software Foundation.
 *
 * MPACK is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Lesser General Public License version 3 for more details
 * (a copy is included in the LICENSE file that accompanied this code).
 *
 * You should have received a copy of the GNU Lesser General Public License
 * version 3 along with MPACK.  If not, see
 * <http://www.gnu.org/licenses/lgpl.html>
 * for a copy of the LGPLv3 License.
 *
 ************************************************************************/

#include <mblas.h>
#include <mlapack.h>

/* PI for GMP */
/* arctan(1/2) + arctan(1/3) = pi/4 */
void
mpf_pi(mpf_t ans)
{
    mpf_t at12, at13, old_at12, old_at13, tmp_at12, tmp_at13, tmp;
    unsigned long times;
    unsigned long prec;

    /*
     * get prec 
     */
    prec = mpf_get_prec(ans);

    /*
     * initialize temporary variables 
     */
    mpf_init2(at12, prec);
    mpf_init2(at13, prec);
    mpf_init2(old_at12, prec);
    mpf_init2(old_at13, prec);
    mpf_init2(tmp_at12, prec);
    mpf_init2(tmp_at13, prec);
    mpf_init2(tmp, prec);

    mpf_set_ui(at12, 1UL);
    mpf_set_ui(at13, 1UL);
    mpf_div_ui(at12, at12, 2UL);	/* at12 = 1/2 */
    mpf_div_ui(at13, at13, 3UL);	/* at13 = 1/3 */

    times = 3;
    mpf_set(tmp_at12, at12);
    mpf_set(old_at12, at12);
    mpf_set(tmp_at13, at13);
    mpf_set(old_at13, at13);

    do {
	mpf_neg(tmp_at12, tmp_at12);
	mpf_div_ui(tmp_at12, tmp_at12, 4UL);
	mpf_div_ui(tmp, tmp_at12, times);
	mpf_add(at12, at12, tmp);

	mpf_neg(tmp_at13, tmp_at13);
	mpf_div_ui(tmp_at13, tmp_at13, 9UL);
	mpf_div_ui(tmp, tmp_at13, times);
	mpf_add(at13, at13, tmp);

	if ((mpf_cmp(at12, old_at12) == 0) && (mpf_cmp(at13, old_at13) == 0))
	    break;

	mpf_set(old_at12, at12);
	mpf_set(old_at13, at13);
	times += 2;
    } while (1);

    mpf_add(ans, at12, at13);
    mpf_mul_ui(ans, ans, 4UL);

    /*
     * clear 
     */
    mpf_clear(at12);
    mpf_clear(at13);
    mpf_clear(old_at12);
    mpf_clear(old_at13);
    mpf_clear(tmp_at12);
    mpf_clear(tmp_at13);
    mpf_clear(tmp);
}

/* exp(1) for GMP */
void
mpf_e(mpf_t ans)
{
    mpf_t old_ans, xn;
    unsigned long times;
    unsigned long prec;

    /*
     * get prec 
     */
    prec = mpf_get_prec(ans);

    /*
     * initialize temporary variables 
     */
    mpf_init2(old_ans, prec);
    mpf_init2(xn, prec);

    /*
     * ans := 1 
     */
    /*
     * old_ans := ans 
     */
    /*
     * xn  := 1   
     */
    /*
     * times := 0 
     */
    mpf_set_ui(ans, 1UL);
    mpf_set(old_ans, ans);
    mpf_set_ui(xn, 1);

    times = 0;

    do {
	/*
	 * /n! 
	 */
	mpf_div_ui(xn, xn, ++times);

	/*
	 * ans += ans + x^n/n! 
	 */
	mpf_add(ans, ans, xn);

	/*
	 * ond_ans == ans 
	 */
	if (mpf_cmp(old_ans, ans) == 0)
	    break;

	/*
	 * old_ans := ans 
	 */
	mpf_set(old_ans, ans);
	//printf(" %ld ", times/2);
    } while (1);

    /*
     * clear 
     */
    mpf_clear(old_ans);
    mpf_clear(xn);
}


/* sin x for GMP */
void
mpf_sin(mpf_t ans, mpf_t x)
{
    mpf_t old_ans, xn, x2, pi2, x_int;
    mp_exp_t exponent;
    char *str;
    long times, sign = 0;	/* 1:plus, -1:minus */
    unsigned long int prec;

    /*
     * get prec 
     */
    prec = mpf_get_prec(ans);

    /*
     * initialize temporary variables 
     */
    mpf_init2(old_ans, prec);
    mpf_init2(xn, prec);
    mpf_init2(x2, prec);
    mpf_init2(pi2, prec);
    mpf_init2(x_int, prec);

    /*
     * change argument 
     */
    mpf_set(xn, x);
    if (mpf_sgn(xn) < 0) {
	sign = -1;		/* sin(-x) = -sin(x) */
	mpf_neg(xn, xn);
    } else
	sign = 1;		/* sin(x) */
    mpf_pi(pi2);
    mpf_mul_ui(pi2, pi2, 2L);	/* pi2 := 2 * PI */
    if (mpf_cmp(xn, pi2) >= 0) {
	mpf_div(x2, xn, pi2);	/* x2 := x - [x/pi2]*pi2 */
	mpf_floor(x2, x2);
	mpf_mul(x2, x2, pi2);
	mpf_sub(xn, xn, x2);
    }
    mpf_div_ui(pi2, pi2, 2L);	/* pi2 := PI */
    if (mpf_cmp(xn, pi2) >= 0) {
	mpf_sub(xn, xn, pi2);	/* x := x - PI */
	sign *= -1;		/* sin(x + PI) = -sin(x) */
    }

    /*
     * ans := changed x   
     */
    /*
     * old_ans := ans 
     */
    /*
     * xn  := x   
     */
    /*
     * x2  := x^2 
     */
    /*
     * times := 1 
     */
    mpf_set(ans, xn);
    mpf_set(old_ans, ans);
    mpf_mul(x2, xn, xn);

    times = 1;

    do {
	/*
	 * x^(2n + 1) 
	 */
	mpf_mul(xn, xn, x2);
	/*
	 * /(2n + 1)! 
	 */
	mpf_div_ui(xn, xn, ++times);
	mpf_div_ui(xn, xn, ++times);
	/*
	 * (-1)^(n+1) 
	 */
	mpf_neg(xn, xn);

	/*
	 * ans += ans + (-1)^(n+1)*x^n/(2*n+1) 
	 */
	mpf_add(ans, ans, xn);

	/*
	 * ond_ans == ans 
	 */
	if (mpf_cmp(old_ans, ans) == 0)
	    break;

	/*
	 * old_ans := ans 
	 */
	mpf_set(old_ans, ans);
	//printf(" %ld ", times/2);
    } while (1);

    /*
     * check sign 
     */
    if (sign == -1)
	mpf_neg(ans, ans);

    /*
     * clear 
     */
    mpf_clear(old_ans);
    mpf_clear(xn);
    mpf_clear(x2);
    mpf_clear(pi2);
    mpf_clear(x_int);

}

/* cos x for GMP */
void
mpf_cos(mpf_t ans, mpf_t x)
{
    mpf_t old_ans, xn, x2, pi2;
    unsigned long times;
    unsigned long int prec;
    long int sign;

    /*
     * get prec 
     */
    prec = mpf_get_prec(ans);

    /*
     * initialize temporary variables 
     */
    mpf_init2(old_ans, prec);
    mpf_init2(xn, prec);
    mpf_init2(x2, prec);
    mpf_init2(pi2, prec);

    /*
     * change argument 
     */
    mpf_set(xn, x);
    sign = 1;
    if (mpf_sgn(xn) < 0) {
	sign = 1;		/* cos(-x) = cos(x) */
	mpf_neg(xn, xn);
    }

    mpf_pi(pi2);
    mpf_mul_ui(pi2, pi2, 2L);	/* pi2 := 2 * PI */
    if (mpf_cmp(xn, pi2) >= 0) {
	mpf_div(x2, xn, pi2);	/* x2 := x - [x/pi2]*pi2 */
	mpf_floor(x2, x2);
	mpf_mul(x2, x2, pi2);
	mpf_sub(xn, xn, x2);
    }
    mpf_div_ui(pi2, pi2, 2L);	/* pi2 := PI */
    if (mpf_cmp(xn, pi2) >= 0) {
	mpf_sub(xn, xn, pi2);	/* x := x - PI */
	sign *= -1;		/* cos(x + PI) = -cos(x) */
    }

    /*
     * ans := 1   
     */
    /*
     * old_ans := ans 
     */
    /*
     * xn  := 1   
     */
    /*
     * x2  := x^2 
     */
    /*
     * times := 0 
     */
    mpf_set_ui(ans, 1UL);
    mpf_set(old_ans, ans);
    mpf_set(x2, xn);
    mpf_mul(x2, x2, xn);
    mpf_set_si(xn, 1L);

    times = 0;

    do {
	/*
	 * x^(2n) 
	 */
	mpf_mul(xn, xn, x2);
	/*
	 * /(2n)! 
	 */
	mpf_div_ui(xn, xn, ++times);
	mpf_div_ui(xn, xn, ++times);
	/*
	 * (-1)^(n+1) 
	 */
	mpf_neg(xn, xn);

	/*
	 * ans += ans + (-1)^(n+1)*x^n/(2*n) 
	 */
	mpf_add(ans, ans, xn);

	/*
	 * ond_ans == ans 
	 */
	if (mpf_cmp(old_ans, ans) == 0)
	    break;

	/*
	 * old_ans := ans 
	 */
	mpf_set(old_ans, ans);
	//printf(" %ld ", times/2);
    } while (1);

    /*
     * check sign 
     */
    if (sign == -1)
	mpf_neg(ans, ans);

    /*
     * clear 
     */
    mpf_clear(old_ans);
    mpf_clear(xn);
    mpf_clear(x2);
}

/* exp(x) for GMP */
void
mpf_exp(mpf_t ans, mpf_t x)
{
    mpf_t old_ans, xn, new_x;
    long int sign;
    unsigned long times;
    unsigned long int prec;

    /*
     * get prec 
     */
    prec = mpf_get_prec(ans);

    /*
     * initialize temporary variables 
     */
    mpf_init2(old_ans, prec);
    mpf_init2(xn, prec);
    mpf_init2(new_x, prec);

    /*
     * x < 0 
     */

    if (mpf_sgn(x) < 0) {
	sign = 0;		/* minus */
	mpf_neg(new_x, x);
    } else {
	sign = 1;		/* plus */
	mpf_set(new_x, x);
    }

    /*
     * ans := 1 
     */
    /*
     * old_ans := ans 
     */
    /*
     * xn  := x   
     */
    /*
     * times := 0 
     */
    mpf_set_ui(ans, 1UL);
    mpf_set(old_ans, ans);
    mpf_set_ui(xn, 1UL);

    times = 0;

    do {
	/*
	 * x^n 
	 */
	mpf_mul(xn, xn, new_x);
	/*
	 * /n! 
	 */
	mpf_div_ui(xn, xn, ++times);

	/*
	 * ans += ans + x^n/n! 
	 */
	mpf_add(ans, ans, xn);

	/*
	 * ond_ans == ans 
	 */
	if (mpf_cmp(old_ans, ans) == 0)
	    break;

	/*
	 * old_ans := ans 
	 */
	mpf_set(old_ans, ans);
	//printf(" %ld ", times/2);
    } while (1);

    if (sign == 0)		/* minus */
	mpf_ui_div(ans, 1UL, ans);

    /*
     * clear 
     */
    mpf_clear(old_ans);
    mpf_clear(xn);
    mpf_clear(new_x);

}

/* log_e 2                           */
/* original code by Haruhiko Okumura */
/* adapted for GMP by Tomonori Kouya */
void
mpf_ln_2(mpf_t ans)
{
    long int i;
    unsigned long int prec;
    mpf_t x, x2, s, last;

    /*
     * get prec of ans 
     */
    prec = mpf_get_prec(ans);

    mpf_init2(x, prec);
    mpf_init2(x2, prec);
    mpf_init2(s, prec);
    mpf_init2(last, prec);

    mpf_set_ui(x, 2UL);

    /*
     * x = (2 - 1) / (2 + 1) 
     */
    mpf_set_ui(x, 1UL);
    mpf_div_ui(x, x, 3UL);

    mpf_mul(x2, x, x);
    i = 1;
    mpf_set(s, x);

    do {
	mpf_mul(x, x, x2);
	i += 2;
	mpf_set(last, s);
	mpf_div_ui(s, x, i);
	mpf_add(s, last, s);
    } while (mpf_cmp(s, last) != 0);

    /*
     * s *= 2; 
     */
    mpf_mul_2exp(ans, s, 1UL);

    mpf_clear(s);
    mpf_clear(x);
    mpf_clear(x2);
    mpf_clear(last);

}

/* log_e x                           */
/* original code by Haruhiko Okumura */
/* adapted for GMP by Tomonori Kouya */
void
mpf_ln(mpf_t ans, mpf_t x)
{
    long int i, k;
    unsigned long int prec;
    mpf_t new_x, x2, s, last, ln2;

    if (mpf_sgn(x) <= 0) {
	fprintf(stderr, "mpf_ln(x): illegal argument\n");
	return;
    }

    /*
     * get prec of ans 
     */
    prec = mpf_get_prec(ans);

    mpf_init2(x2, prec);
    mpf_init2(new_x, prec);
    mpf_init2(s, prec);
    mpf_init2(last, prec);
    mpf_init2(ln2, prec);

    /*
     * get exponent 
     */
    mpf_set(new_x, x);
    k = new_x->_mp_exp;

    if (k < 0)
	mpf_mul_2exp(new_x, new_x, (unsigned long)(-k));
    else
	mpf_div_2exp(new_x, new_x, (unsigned long)k);

    /*
     * x = (x - 1) / (x + 1) 
     */
    mpf_sub_ui(new_x, new_x, 1UL);
    mpf_add_ui(x2, new_x, 2UL);
    mpf_div(new_x, new_x, x2);

    mpf_mul(x2, new_x, new_x);
    i = 1;
    mpf_set(s, new_x);

    do {
	mpf_mul(new_x, new_x, x2);
	i += 2;
	mpf_set(last, s);
	mpf_div_ui(s, new_x, i);
	mpf_add(s, last, s);
    } while (mpf_cmp(s, last) != 0);

    /*
     * s *= 2; 
     */
    mpf_mul_2exp(ans, s, 1UL);
    mpf_ln_2(ln2);
    if (k < 0) {
	k = -k;
	mpf_mul_ui(x2, ln2, k);
	mpf_neg(x2, x2);
    } else
	mpf_mul_ui(x2, ln2, k);
    mpf_add(ans, ans, x2);

    mpf_clear(s);
    mpf_clear(new_x);
    mpf_clear(x2);
    mpf_clear(last);
    mpf_clear(ln2);

}

/* log_10 x                          */
void
mpf_log10(mpf_t ans, mpf_t x)
{
    unsigned long int prec;
    mpf_t ln10;

    if (mpf_sgn(x) <= 0) {
	fprintf(stderr, "mpf_log10(x): illegal argument\n");
	return;
    }

    /*
     * init 
     */
    prec = mpf_get_prec(ans);
    mpf_init2(ln10, prec);

    /*
     * log10 x := ln x / ln 10 
     */
    mpf_ln(ans, x);
    mpf_set_ui(ln10, 10UL);
    mpf_ln(ln10, ln10);
    mpf_div(ans, ans, ln10);

    /*
     * clear 
     */
    mpf_clear(ln10);

    return;
}
